import abc
import time
import functools
import isddc as i

from . import itime
from . import isddc_log as ilog


class BaseDecorator(metaclass=abc.ABCMeta):

    def __call__(self, func):
        @functools.wraps(func)
        def inner(*a, **kw):
            return self.core(func, *a, **kw)
        return inner

    @abc.abstractmethod
    def core(self, func, *a, **kw):
        ...

    @staticmethod
    def output(msg: str, *, level: str = 'ERROR'):
        try:
            getattr(ilog.simple, level.lower())(msg)
        except AttributeError:
            return print(f'[{itime.now.str}] [{level.upper()}] {msg}')


class TryExcept(BaseDecorator):

    def __init__(
            self,
            exc_type: type or tuple = Exception,
            *,
            mark: str = None,
            ignore: bool = False,
            exc_return=None,
            exc_callback=None,
    ):
        self.mark = mark
        self.ignore = ignore
        self.exc_type = exc_type
        self.exc_return = exc_return
        self.exc_callback = exc_callback

    def core(self, func, *a, **kw):
        try:
            return func(*a, **kw)
        except self.exc_type as e:
            if self.exc_callback:
                return self.exc_callback(*a, **kw)

            if not self.ignore:
                try:
                    self.exception_handler(func.__name__, e)
                except Exception as e:
                    self.output(f'[{itime.now.str}] TryExceptError: {e}')

        return self.exc_return

    def exception_handler(self, func_name: str, err: Exception):
        mark: str = self.mark or i.hump(func_name)
        exc_name: str = type(err).__name__
        self.output(f'{mark}.{exc_name}: {str(err)}')


class Retry(BaseDecorator):

    def __init__(
            self,
            mark: str = None,
            *,
            count: int = None,
            cycle: int = 0,
            retry_exc: type or tuple = Exception,
    ):
        self.mark = mark
        self.count = count
        self.cycle = cycle
        self.retry_exc = retry_exc

    def core(self, func, *a, **kw):
        count = 0

        while True:
            try:
                return func(*a, **kw)
            except self.retry_exc as e:
                try:
                    count += 1
                    sign: str = self.mark or i.hump(func.__name__)
                    exc_name: str = type(e).__name__
                    self.output(
                        f'[attempts:{count}/{self.count or "N"}] {sign}.{exc_name}: {e}',
                        level='WARNING')
                except Exception as ee:
                    self.output(f'[{itime.now.str}] RetryError: {ee}')

                if self.count and count == self.count:
                    raise e

            time.sleep(self.cycle)


class WhileTrue(BaseDecorator):

    def __init__(
            self,
            cond=True,
            *,
            cycle: int = 0,
            precise: bool = False,
            before: bool = False
    ):
        self.cond = cond
        self.cycle = cycle
        self.before = before
        self.core = self.precise if precise else self.simple

    def core(self, func, *a, **kw):
        pass

    def simple(self, func, *a, **kw):
        while self.cond:
            self.before and time.sleep(self.cycle)
            ret = func(*a, **kw)
            if ret == '_break_':
                break
            if ret == '_continue_':
                continue
            self.before or time.sleep(self.cycle)

    def precise(self, func, *a, **kw):
        cycle = self.cycle
        while self.cond:
            self.before and time.sleep(cycle)
            start = time.time()
            ret = func(*a, **kw)
            if ret == '_break_':
                break
            if ret == '_continue_':
                continue
            end = time.time()
            cycle: float = abs(self.cycle - (end - start))
            self.before or time.sleep(cycle)


class RunTime(BaseDecorator):

    def __init__(self, mark: str = None):
        self.mark = mark

    def core(self, func, *a, **kw):
        start = time.time()
        func(*a, **kw)
        end = time.time()

        mark: str = self.mark or i.hump(func.__name__)
        exec_time: str = itime.second2time(end - start)
        self.output(f'{mark}: {exec_time}', level='INFO')


class TestFuncSpeed(BaseDecorator):

    def __init__(self, count: int, keep: int = 7):
        self.count = count
        self.keep = keep

    def core(self, func, *a, **kw):
        speeds = []

        for _ in range(self.count):
            start = time.time()
            func(*a, **kw)
            end = time.time()
            speeds.append(end - start)

        for action in min, max, i.avg, sum:
            result = round(action(speeds), self.keep)
            print(f'{action.__name__}: {result}')


class BeforeFunc(BaseDecorator):

    def __init__(self, before_func, *, enable: bool = True):
        self.before_func = before_func
        self.enable = enable

    def core(self, func, *a, **kw):
        if not self.enable:
            return func(*a, **kw)

        try:
            result = self.before_func(*a, **kw)
        except TypeError:
            result = self.before_func()

        if not isinstance(result, dict):
            return func(*a, **kw)

        a = list(a)
        co_varnames: tuple = func.__code__.co_varnames

        if co_varnames == ('a', 'kw'):
            return func(*a, **kw)

        for idx, param in enumerate(co_varnames):
            if param in result:
                try:
                    a[idx] = result[param]
                except IndexError:
                    kw[param] = result[param]

        return func(*a, **kw)


class AfterFunc(BaseDecorator):

    def __init__(
            self,
            after_func,
            *,
            use_raw_params: bool = False,
            enable: bool = True
    ):
        self.after_func = after_func
        self.use_raw_params = use_raw_params
        self.enable = enable

    def core(self, func, *a, **kw):
        raw_result = func(*a, **kw)

        if not self.enable:
            return raw_result

        if not self.after_func.__code__.co_varnames:
            result = self.after_func()
        elif self.use_raw_params:
            result = self.after_func(*a, **kw)
        else:
            try:
                result = self.after_func(a[0], raw_result)
            except (IndexError, TypeError):
                result = self.after_func(raw_result)

        return raw_result if result is None else result
